==> eslint-plugin-total-functions-1.25.0.tgz: package/src/rules/no-array-destructuring.ts <==
import { RuleModule } from "@typescript-eslint/experimental-utils/dist/ts-eslint";
import {
  ESLintUtils,
  AST_NODE_TYPES,
} from "@typescript-eslint/experimental-utils";
import { isTupleTypeReference } from "tsutils";
import ts from "typescript";

/**
 * An ESLint rule to ban unsafe array/object destructuring, which is not well-typed in TypeScript.
 */
const noArrayDestructuring: RuleModule<"errorStringGeneric", readonly []> = {
  meta: {
    type: "problem",
    docs: {
      category: "Possible Errors",
      description:
        "Array and object destructuring are not type-safe in TypeScript.",
      recommended: "error",
      url: "https://github.com/danielnixon/eslint-plugin-total-functions",
    },
    messages: {
      errorStringGeneric: "This destructuring is not type-safe in TypeScript.",
    },
    schema: [],
  },
  create: (context) => {
    // eslint-disable-next-line total-functions/no-unsafe-assignment
    const parserServices = ESLintUtils.getParserServices(context);
    const checker = parserServices.program.getTypeChecker();

    return {
      // eslint-disable-next-line functional/no-return-void
      ArrayPattern: (node): void => {
        const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
        const type = checker.getTypeAtLocation(tsNode);

        // eslint-disable-next-line functional/no-conditional-statement
        if (
          isTupleTypeReference(type) &&
          node.elements.length <= type.target.minLength
        ) {
          // Allow tuples (whether partial or not) if the indexes
          // we're destructuring are all know to be safe at compile time.
          return;
        }

        const numberIndexType = type.getNumberIndexType();
        // eslint-disable-next-line functional/no-conditional-statement
        if (numberIndexType !== undefined) {
          const typeParts = numberIndexType.isUnion()
            ? numberIndexType.types
            : [numberIndexType];
          // eslint-disable-next-line functional/no-conditional-statement
          if (
            typeParts.find((t) => t.flags & ts.TypeFlags.Undefined) !==
            undefined
          ) {
            // Allow destructuring if undefined is already in the array type.
            return;
          }
        }

        // eslint-disable-next-line functional/no-expression-statement
        context.report({
          node: node,
          messageId: "errorStringGeneric",
        });
      },
      // eslint-disable-next-line functional/no-return-void
      ObjectPattern: (node): void => {
        const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
        const type = checker.getTypeAtLocation(tsNode);

        // True if every element we're destructuring is known to exist at compile time (or is the rest element).
        const isSafe = node.properties.every(
          (prop) =>
            prop.type === AST_NODE_TYPES.RestElement ||
            (prop.key.type === AST_NODE_TYPES.Identifier &&
              type.getProperty(prop.key.name) !== undefined)
        );

        // eslint-disable-next-line functional/no-conditional-statement
        if (isSafe) {
          return;
        }

        // eslint-disable-next-line functional/no-expression-statement
        context.report({
          node: node,
          messageId: "errorStringGeneric",
        });
      },
    };
  },
};

export default noArrayDestructuring;

==> eslint-plugin-total-functions-1.29.0.tgz: package/src/rules/no-array-destructuring.ts <==
import { RuleModule } from "@typescript-eslint/experimental-utils/dist/ts-eslint";
import {
  ESLintUtils,
  AST_NODE_TYPES,
} from "@typescript-eslint/experimental-utils";
import { isTupleTypeReference } from "tsutils";
import ts from "typescript";

/**
 * An ESLint rule to ban unsafe array/object destructuring, which is not well-typed in TypeScript.
 */
const noArrayDestructuring: RuleModule<"errorStringGeneric", readonly []> = {
  meta: {
    type: "problem",
    docs: {
      category: "Possible Errors",
      description:
        "Array and object destructuring are not type-safe in TypeScript.",
      recommended: "error",
      url: "https://github.com/danielnixon/eslint-plugin-total-functions",
    },
    messages: {
      errorStringGeneric: "This destructuring is not type-safe in TypeScript.",
    },
    schema: [],
  },
  create: (context) => {
    const parserServices = ESLintUtils.getParserServices(context);
    const checker = parserServices.program.getTypeChecker();

    return {
      // eslint-disable-next-line functional/no-return-void
      ArrayPattern: (node): void => {
        const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
        const type = checker.getTypeAtLocation(tsNode);

        // eslint-disable-next-line functional/no-conditional-statement
        if (
          isTupleTypeReference(type) &&
          node.elements.length <= type.target.minLength
        ) {
          // Allow tuples (whether partial or not) if the indexes
          // we're destructuring are all know to be safe at compile time.
          return;
        }

        const numberIndexType = type.getNumberIndexType();
        const typeParts =
          numberIndexType !== undefined
            ? numberIndexType.isUnion()
              ? numberIndexType.types
              : [numberIndexType]
            : [];

        // eslint-disable-next-line functional/no-conditional-statement
        if (
          typeParts.find((t) => t.flags & ts.TypeFlags.Undefined) !== undefined
        ) {
          // Allow destructuring if undefined is already in the array type.
          return;
        }

        // eslint-disable-next-line functional/no-expression-statement
        context.report({
          node: node,
          messageId: "errorStringGeneric",
        });
      },
      // eslint-disable-next-line functional/no-return-void
      ObjectPattern: (node): void => {
        const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
        const type = checker.getTypeAtLocation(tsNode);

        // True if every element we're destructuring is known to exist at compile time (or is the rest element).
        const isSafe = node.properties.every(
          (prop) =>
            prop.type === AST_NODE_TYPES.RestElement ||
            (prop.key.type === AST_NODE_TYPES.Identifier &&
              type.getProperty(prop.key.name) !== undefined)
        );

        // eslint-disable-next-line functional/no-conditional-statement
        if (isSafe) {
          return;
        }

        // eslint-disable-next-line functional/no-expression-statement
        context.report({
          node: node,
          messageId: "errorStringGeneric",
        });
      },
    };
  },
};

export default noArrayDestructuring;
